try:
    import psycopg
except ImportError:
    try:
        import psycopg2 as psycopg
    except ImportError:
        try:
            import psycopg2cffi as psycopg
        except ImportError:
            psycopg = None

try:
    from psycopg.conninfo import make_conninfo
except ImportError:
    try:
        from psycopg2.extensions import make_dsn as make_conninfo
    except ImportError:
        make_conninfo = lambda value: value  # noqa



import pytest


def pytest_collection_modifyitems(items):
    for item in items:
        if item.name.startswith("test_func"):
            item.add_marker(pytest.mark.func)
        else:
            item.add_marker(pytest.mark.unit)



"""
Module that contains the command line app.

Why does this file exist, and why not put this in __main__?

  You might be tempted to import things from __main__ later, but that will cause
  problems: the code will get executed twice:

  - When you run `python -mholdup` python will execute
    ``__main__.py`` as a script. That means there will not be any
    ``holdup.__main__`` in ``sys.modules``.
  - When you import __main__ it will get executed again (as a module) because
    there"s no ``holdup.__main__`` in ``sys.modules``.

  Also see (1) from http://click.pocoo.org/5/setuptools/#setuptools-integration
"""

import argparse
import os
import sys
from operator import methodcaller
from shlex import quote
from time import sleep
from time import time

from . import __version__
from .checks import AnyCheck
from .checks import EvalCheck
from .checks import HttpCheck
from .checks import PathCheck
from .checks import PgCheck
from .checks import TcpCheck
from .checks import UnixCheck
from .pg import make_conninfo
from .pg import psycopg


def parse_service(service):
    if "://" not in service:
        raise argparse.ArgumentTypeError(f'Invalid service spec {service!r}. Must have "://".')
    proto, value = service.split("://", 1)

    if "," in value and proto != "eval":
        parts = value.split(",")
        for pos, part in enumerate(parts):
            if part.startswith("eval://"):
                parts[pos] = ",".join(parts[pos:])
                del parts[pos + 1 :]
                break
        return AnyCheck([parse_value(part, proto) for part in parts])
    else:
        return parse_value(value, proto)


def parse_value(value, proto):
    if "://" in value:
        proto, value = value.split("://", 1)
    display_value = f"{proto}://{value}"

    if proto == "tcp":
        if ":" not in value:
            raise argparse.ArgumentTypeError(f'Invalid service spec {display_value!r}. Must have ":". Where\'s the port?')
        host, port = value.strip("/").split(":", 1)
        if not port.isdigit():
            raise argparse.ArgumentTypeError(f"Invalid service spec {display_value!r}. Port must be a number not {port!r}.")
        port = int(port)
        return TcpCheck(host, port)
    elif proto in ("pg", "postgresql", "postgres"):
        if psycopg is None:
            raise argparse.ArgumentTypeError(f"Protocol {proto} unusable. Install holdup[pg].")

        uri = f"postgresql://{value}"
        try:
            connection_uri = make_conninfo(uri)
        except Exception as exc:
            raise argparse.ArgumentTypeError(f"Failed to parse {display_value!r}: {exc}. Must be a valid connection URI.") from None
        return PgCheck(connection_uri)
    elif proto == "unix":
        return UnixCheck(value)
    elif proto == "path":
        return PathCheck(value)
    elif proto in ("http", "https", "https+insecure"):
        return HttpCheck(f"{proto}://{value}")
    elif proto == "eval":
        return EvalCheck(value)
    else:
        raise argparse.ArgumentTypeError(f'Unknown protocol {proto!r} in {display_value!r}. Must be "tcp", "path", "unix" or "pg".')


parser = argparse.ArgumentParser(
    usage="%(prog)s [-h] [-t SECONDS] [-T SECONDS] [-i SECONDS] [-n] service [service ...] " "[-- command [arg [arg ...]]]",
    description="Wait for services to be ready and optionally exec command.",
)
parser.add_argument(
    "service",
    nargs=argparse.ONE_OR_MORE,
    type=parse_service,
    help="A service to wait for. "
    "Supported protocols: "
    '"tcp://host:port/", '
    '"path:///path/to/something", '
    '"unix:///path/to/domain.sock", '
    '"eval://expr", '
    '"pg://user:password@host:port/dbname" ("postgres" and "postgresql" also allowed), '
    '"http://urn", '
    '"https://urn", '
    '"https+insecure://urn" (status 200 expected for http*). '
    "Join protocols with a comma to make holdup exit at the first "
    'passing one, eg: "tcp://host:1,host:2" or "tcp://host:1,tcp://host:2" are equivalent and mean '
    "`any that pass`.",
)
parser.add_argument("command", nargs=argparse.OPTIONAL, help="An optional command to exec.")
parser.add_argument(
    "-t", "--timeout", metavar="SECONDS", type=float, default=60.0, help="Time to wait for services to be ready. Default: %(default)s"
)
parser.add_argument(
    "-T", "--check-timeout", metavar="SECONDS", type=float, default=1.0, help="Time to wait for a single check. Default: %(default)s"
)
parser.add_argument("-i", "--interval", metavar="SECONDS", type=float, default=0.2, help="How often to check. Default: %(default)s")
parser.add_argument("-v", "--verbose", action="store_true", help="Verbose mode.")
parser.add_argument("--verbose-passwords", action="store_true", help="Disable PostgreSQL/HTTP password masking.")
parser.add_argument(
    "-n",
    "--no-abort",
    action="store_true",
    help="Ignore failed services. " "This makes `holdup` return 0 exit code regardless of services actually responding.",
)
parser.add_argument("--insecure", action="store_true", help="Disable SSL Certificate verification for HTTPS services.")


def add_version_argument(parser):
    parser.add_argument(
        "--version",
        action="version",
        version=f"%(prog)s v{__version__}",
        help="display the version of the holdup package and its location, then exit.",
    )


def main():
    """
    Args:
        argv (list): List of arguments

    Returns:
        int: A return code

    Does stuff.
    """
    add_version_argument(parser)
    if "--" in sys.argv:
        pos = sys.argv.index("--")
        argv, command = sys.argv[1:pos], sys.argv[pos + 1 :]
    else:
        argv, command = sys.argv[1:], None
    options = parser.parse_args(args=argv)
    if options.timeout < options.check_timeout:
        if options.check_timeout == 1.0:
            options.check_timeout = options.timeout
        else:
            parser.error("--timeout value must be greater than --check-timeout value!")
    pending = list(options.service)
    brief_representer = methodcaller("display", verbose=False, verbose_passwords=options.verbose_passwords)
    verbose_representer = methodcaller("display", verbose=True, verbose_passwords=options.verbose_passwords)
    if options.verbose:
        print(
            f"holdup: Waiting for {options.timeout}s ({options.check_timeout}s per check, {options.interval}s sleep between loops) "
            f'for these services: {", ".join(map(brief_representer, pending))}'
        )
    start = time()
    at_least_once = True
    while at_least_once or pending and time() - start < options.timeout:
        lapse = time()
        pending = [check for check in pending if not check.is_passing(options)]
        sleep(max(0, options.interval - time() + lapse))
        at_least_once = False

    if pending:
        if options.no_abort:
            print(
                f'holdup: Failed checks: {", ".join(map(verbose_representer, pending))}. Treating as success because of --no-abort.',
                file=sys.stderr,
            )
        else:
            parser.exit(1, f'holdup: Failed checks: {", ".join(map(verbose_representer, pending))}. Aborting!\n')
    if command:
        if options.verbose:
            print(f'holdup: Executing: {" ".join(map(quote, command))}')
        os.execvp(command[0], command)  # noqa:S606



import argparse
import ast
import builtins
import os
import re
import socket
import ssl
import sys
from contextlib import closing
from operator import methodcaller
from urllib.parse import urlparse
from urllib.parse import urlunparse
from urllib.request import HTTPBasicAuthHandler
from urllib.request import HTTPDigestAuthHandler
from urllib.request import HTTPPasswordMgrWithDefaultRealm
from urllib.request import HTTPSHandler
from urllib.request import Request
from urllib.request import build_opener

from . import __version__
from .pg import psycopg


class Check:
    error = None

    def is_passing(self, options):
        try:
            self.run(options)
        except Exception as exc:
            self.error = exc
        else:
            self.error = False
            if options.verbose:
                print(f"holdup: Passed check: {self.display(verbose=True, verbose_passwords=options.verbose_passwords)}")
            return True

    def run(self, options):
        raise NotImplementedError

    @property
    def status(self):
        if self.error:
            return f"{self.error}"
        elif self.error is None:
            return "PENDING"
        else:
            return "PASSED"

    def __repr__(self):
        return f"{self.__class__.__name__}({repr(self.__dict__)[1:-1]})"

    def display_definition(self, **kwargs):
        raise NotImplementedError

    def display(self, *, verbose, **kwargs):
        definition = self.display_definition(**kwargs)
        if verbose:
            return f"{definition!r} -> {self.status}"
        else:
            return definition


class TcpCheck(Check):
    def __init__(self, host, port):
        self.host = host
        self.port = port

    def run(self, options):
        sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        sock.settimeout(options.check_timeout)
        with closing(sock):
            sock.connect((self.host, self.port))

    def __repr__(self):
        return f"TcpCheck(host={self.host!r}, port={self.port!r})"

    def display(self, *, verbose, **_):
        definition = f"tcp://{self.host}:{self.port}"
        if verbose:
            return f"{definition!r} -> {self.status}"
        else:
            return definition


class PgCheck(Check):
    def __init__(self, connection_string):
        self.connection_string = connection_string
        if "?" in connection_string.rsplit("/", 1)[1]:
            self.separator = "&"
        else:
            self.separator = "?"

    def run(self, options):
        with closing(
            psycopg.connect(f"{self.connection_string}{self.separator}connect_timeout={max(1, int(options.check_timeout))}")
        ) as conn:
            with closing(conn.cursor()) as cur:
                cur.execute("SELECT version()")
                cur.fetchone()

    def __repr__(self):
        return f"PgCheck({self.connection_string})"

    def display_definition(self, *, verbose_passwords, _password_re=re.compile(r":[^@:]+@")):
        definition = str(self.connection_string)
        if not verbose_passwords:
            definition = _password_re.sub(":******@", definition, 1)
        return definition


class HttpCheck(Check):
    def __init__(self, url):
        self.handlers = []
        self.parsed_url = url = urlparse(url)
        self.scheme = url.scheme
        self.insecure = False
        if url.scheme == "https+insecure":
            self.insecure = True
            url = url._replace(scheme="https")

        if url.port:
            self.netloc = f"{url.hostname}:{url.port}"
        else:
            self.netloc = url.hostname
        self.host = url.hostname

        cleaned_url = urlunparse(url._replace(netloc=self.netloc))

        if url.username or url.password:
            password_mgr = HTTPPasswordMgrWithDefaultRealm()
            password_mgr.add_password(None, cleaned_url, url.username, url.password)
            self.handlers.append(HTTPDigestAuthHandler(passwd=password_mgr))
            self.handlers.append(HTTPBasicAuthHandler(password_mgr=password_mgr))

        self.url = cleaned_url

    def run(self, options):
        handlers = list(self.handlers)
        insecure = self.insecure or options.insecure

        ssl_ctx = ssl.create_default_context()
        if insecure:
            ssl_ctx.check_hostname = False
            ssl_ctx.verify_mode = ssl.CERT_NONE
        handlers.append(HTTPSHandler(context=ssl_ctx))

        opener = build_opener(*handlers)
        opener.addheaders = [("User-Agent", f"python-holdup/{__version__}")]
        request = Request(self.url, headers={"Host": self.host})  # noqa: S310
        with closing(opener.open(request, timeout=options.check_timeout)) as req:
            status = req.getcode()
            if status != 200:
                raise Exception(f"Expected status code 200, got {status!r}")

    def __repr__(self):
        return f"HttpCheck({self.url}, insecure={self.insecure}, status={self.status})"

    def display_definition(self, *, verbose_passwords):
        url = self.parsed_url
        if url.username and not verbose_passwords:
            if not url.password:
                mask = "******"
            else:
                mask = f"{url.username}:******"
            url = url._replace(netloc=f"{mask}@{self.netloc}")
        return urlunparse(url)


class UnixCheck(Check):
    def __init__(self, path):
        self.path = path

    def run(self, options):
        sock = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)
        sock.settimeout(options.check_timeout)
        with closing(sock):
            sock.connect(self.path)

    def __repr__(self):
        return f"UnixCheck({self.path!r}, status={self.status})"

    def display_definition(self, **_):
        return f"unix://{self.path}"


class PathCheck(Check):
    def __init__(self, path):
        self.path = path

    def run(self, _):
        # necessary to check if it exists.
        os.stat(self.path)  # noqa: PTH116
        if not os.access(self.path, os.R_OK):
            raise Exception(f"Failed access({self.path!r}, R_OK) test")

    def __repr__(self):
        return f"PathCheck({self.path!r}, status={self.status})"

    def display_definition(self, **_):
        return f"path://{self.path}"


class EvalCheck(Check):
    def __init__(self, expr):
        self.expr = expr
        self.ns = {}
        try:
            tree = ast.parse(expr)
        except SyntaxError as exc:
            raise argparse.ArgumentTypeError(
                f'Invalid service spec {expr!r}. Parse error:\n  {exc.text} {" " * exc.offset}^\n{exc}'
            ) from None
        for node in ast.walk(tree):
            if isinstance(node, ast.Name):
                if not hasattr(builtins, node.id):
                    try:
                        __import__(node.id)
                    except ImportError as exc:
                        raise argparse.ArgumentTypeError(f"Invalid service spec {expr!r}. Import error: {exc}") from None
                    self.ns[node.id] = sys.modules[node.id]

    def run(self, _):
        result = eval(self.expr, dict(self.ns), dict(self.ns))  # noqa: S307
        if not result:
            raise Exception(f"Failed to evaluate {self.expr!r}. Result {result!r} is falsey")

    def __repr__(self):
        return f"EvalCheck({self.expr!r}, ns={self.ns!r}, status={self.status})"

    def display_definition(self, **_):
        return f"eval://{self.expr}"


class AnyCheck(Check):
    def __init__(self, checks):
        self.checks = checks

    def run(self, options):
        for check in self.checks:
            if check.is_passing(options):
                break
        else:
            raise Exception("ALL FAILED")

    def __repr__(self):
        return f'AnyCheck({", ".join(map(repr, self.checks))}, status={self.status})'

    def display(self, *, verbose, **kwargs):
        checks = ", ".join(map(methodcaller("display", verbose=verbose, **kwargs), self.checks))
        if verbose:
            return f"any({checks}) -> {self.status}"
        else:
            return f"any({checks})"



# ruff: noqa: PTH110, PTH120, PTH123
import os
import platform
import shutil
import socket
import threading

import pytest

pytest_plugins = ("pytester",)


def has_docker():
    return shutil.which("docker")


@pytest.fixture(params=[[], ["--", "python", "-c", 'print("success !")']])
def extra(request):
    return request.param


def test_normal(testdir, tmp_path_factory, extra):
    tcp = socket.socket()
    tcp.bind(("127.0.0.1", 0))
    tcp.listen(1)
    _, port = tcp.getsockname()

    def accept():
        conn, _ = tcp.accept()
        conn.close()

    t = threading.Thread(target=accept)
    t.start()

    uds = socket.socket(socket.AF_UNIX)

    tmp_path = tmp_path_factory.getbasetemp()
    unix_path = tmp_path / "sock"
    path_path = tmp_path / "file"

    with open(path_path, "w"):
        pass
    uds.bind(str(unix_path))
    uds.listen(1)

    result = testdir.run("holdup", "-t", "0.5", f"tcp://localhost:{port}/", f"path://{path_path}", f"unix://{unix_path}", *extra)
    if extra:
        result.stdout.fnmatch_lines(["success !"])
    assert result.ret == 0
    t.join()
    uds.close()
    tcp.close()
    unix_path.unlink()


@pytest.mark.parametrize("status", [200, 404])
@pytest.mark.parametrize("proto", ["http", "https"])
def test_http(testdir, extra, status, proto):
    result = testdir.run("holdup", "-T", "5", "-t", "5.1", f"{proto}://httpbingo.org/status/{status}", *extra)
    if extra:
        if status == 200:
            result.stdout.fnmatch_lines(["success !"])
        else:
            result.stderr.fnmatch_lines(["*HTTP Error 404*"])


@pytest.mark.parametrize("status", [200, 404])
def test_http_port(testdir, extra, status):
    result = testdir.run("holdup", "-T", "5", "-t", "5.1", f"http://httpbingo.org:80/status/{status}", *extra)
    if extra:
        if status == 200:
            result.stdout.fnmatch_lines(["success !"])
        else:
            result.stderr.fnmatch_lines(["*HTTP Error 404*"])


@pytest.mark.parametrize("auth", ["basic-auth", "digest-auth/auth"])
@pytest.mark.parametrize("proto", ["http", "https"])
def test_http_auth(testdir, extra, auth, proto):
    result = testdir.run("holdup", "-T", "5", "-t", "5.1", f"{proto}://usr:pwd@httpbingo.org/{auth}/usr/pwd", *extra)
    if extra:
        result.stdout.fnmatch_lines(["success !"])


def test_http_insecure_with_option(testdir):
    result = testdir.run("holdup", "-t", "2", "--insecure", "https://self-signed.badssl.com/")
    assert result.ret == 0


def test_http_insecure_with_proto(testdir):
    result = testdir.run("holdup", "-t", "2", "https+insecure://self-signed.badssl.com/")
    assert result.ret == 0


def test_any1(testdir, tmp_path_factory, extra):
    tcp = socket.socket()
    tcp.bind(("127.0.0.1", 0))
    _, port = tcp.getsockname()

    uds = socket.socket(socket.AF_UNIX)

    tmp_path = tmp_path_factory.getbasetemp()
    unix_path = tmp_path / "s"
    path_path = tmp_path / "miss"
    uds.bind(str(unix_path))
    uds.listen(1)
    result = testdir.run("holdup", "-v", "-t", "0.5", f"tcp://localhost:{port}/,path://{path_path},unix://{unix_path}", *extra)
    if extra:
        result.stdout.fnmatch_lines(
            [
                "holdup: Waiting for 0.5s (0.5s per check, 0.2s sleep between loops) for these services: "
                f"any(tcp://localhost:*, path://{path_path}, unix://{unix_path})",
                f"holdup: Passed check: 'unix://{unix_path}' -> PASSED",
                f"holdup: Passed check: any('tcp://localhost:*' -> *, 'path://{path_path}' ->"
                f" [[]Errno 2[]] No such file or directory: *, "
                f"'unix://{unix_path}' -> PASSED) -> PASSED",
                "holdup: Executing: python -c 'print(\"success !\")'",
                "success !",
            ]
        )
    assert result.ret == 0
    tcp.close()
    uds.close()
    unix_path.unlink()


def test_any2(testdir, tmp_path, extra):
    tcp = socket.socket()
    tcp.bind(("127.0.0.1", 0))
    _, port = tcp.getsockname()

    uds = socket.socket(socket.AF_UNIX)
    unix_path = tmp_path / "s"
    path_path = tmp_path / "miss"
    uds.bind(str(unix_path))
    uds.listen(1)
    result = testdir.run("holdup", "-v", "-t", "0.5", f"path://{path_path},unix://{unix_path},tcp://localhost:{port}/", *extra)
    if extra:
        result.stdout.fnmatch_lines(
            [
                "holdup: Waiting for 0.5s (0.5s per check, 0.2s sleep between loops) for these services: "
                f"any(path://{path_path}, unix://{unix_path}, tcp://localhost:*)",
                f"holdup: Passed check: 'unix://{unix_path}' -> PASSED",
                f"holdup: Passed check: any('path://{path_path}' -> [[]Errno 2[]] No such file or directory: *, "
                f"'unix://{unix_path}' -> PASSED, 'tcp://localhost:*' -> PENDING) -> PASSED",
                "holdup: Executing: python -c 'print(\"success !\")'",
                "success !",
            ]
        )
    assert result.ret == 0
    tcp.close()
    uds.close()
    unix_path.unlink()


def test_any_same_proto(testdir, extra):
    tcp1 = socket.socket()
    tcp1.bind(("127.0.0.1", 0))
    _, port1 = tcp1.getsockname()

    tcp2 = socket.socket()
    tcp2.bind(("127.0.0.1", 0))
    tcp2.listen(1)
    _, port2 = tcp2.getsockname()

    def accept():
        conn, _ = tcp2.accept()
        conn.close()

    t = threading.Thread(target=accept)
    t.start()

    result = testdir.run("holdup", "-t", "0.5", f"tcp://localhost:{port1},localhost:{port2}/", *extra)
    if extra:
        result.stdout.fnmatch_lines(["success !"])
    assert result.ret == 0
    t.join()
    tcp1.close()
    tcp2.close()


def test_any_failed(testdir):
    tcp = socket.socket()
    tcp.bind(("127.0.0.1", 0))
    _, port = tcp.getsockname()

    result = testdir.run("holdup", "-t", "0.5", f"tcp://localhost:{port}/,path:///doesnt/exist,unix:///doesnt/exist")
    result.stderr.fnmatch_lines(
        [
            f"holdup: Failed checks: any('tcp://localhost:{port}' -> *, 'path:///doesnt/exist' -> *, "
            "'unix:///doesnt/exist' -> *) -> ALL FAILED. "
            "Aborting!",
        ]
    )
    tcp.close()


def test_no_abort(testdir, extra):
    result = testdir.run(
        "holdup", "-t", "0.1", "-n", "tcp://localhost:0", "tcp://localhost:0/", "path:///doesnt/exist", "unix:///doesnt/exist", *extra
    )
    result.stderr.fnmatch_lines(
        [
            "holdup: Failed checks: 'tcp://localhost:0' -> *, "
            "'path:///doesnt/exist' -> *, 'unix:///doesnt/exist' -> *. "
            "Treating as success because of --no-abort.",
        ]
    )


@pytest.mark.skipif(os.path.exists("/.dockerenv"), reason="chmod(0) does not work in docker")
def test_not_readable(testdir, extra):
    foobar = testdir.maketxtfile(foobar="")
    foobar.chmod(0)
    result = testdir.run("holdup", "-t", "0.1", "-n", f"path://{foobar}", *extra)
    result.stderr.fnmatch_lines(
        [
            f"holdup: Failed checks: 'path://{foobar}' -> Failed access('{foobar}', R_OK) test. "
            "Treating as success because of --no-abort.",
        ]
    )


def test_bad_timeout(testdir):
    result = testdir.run("holdup", "-t", "0.1", "-T", "2", "path:///")
    result.stderr.fnmatch_lines(["*error: --timeout value must be greater than --check-timeout value!"])


def test_eval_bad_import(testdir):
    result = testdir.run("holdup", "eval://foobar123.foo()")
    result.stderr.fnmatch_lines(
        [
            "*error: argument service: Invalid service spec 'foobar123.foo()'. Import error: No module named*",
        ]
    )


def test_eval_bad_expr(testdir):
    result = testdir.run("holdup", "eval://foobar123.foo(.)")
    result.stderr.fnmatch_lines(
        [
            "*error: argument service: Invalid service spec 'foobar123.foo(.)'. Parse error:",
            "  foobar123.foo(.)",
            "*               ^",
            "invalid syntax (<unknown>, line 1)",
        ]
    )


def test_eval_falsey(testdir):
    result = testdir.run("holdup", "-t", "0", "eval://None")
    result.stderr.fnmatch_lines(["holdup: Failed checks: 'eval://None' -> Failed to evaluate 'None'. Result None is falsey. Aborting!"])
    assert result.ret == 1


def test_eval_distutils(testdir, extra):
    result = testdir.run("holdup", 'eval://__import__("shutil").which("find")', *extra)
    if extra:
        result.stdout.fnmatch_lines(["success !"])
    assert result.ret == 0


def test_eval_comma(testdir, extra):
    result = testdir.run("holdup", 'eval://os.path.join("foo", "bar")', *extra)
    if extra:
        result.stdout.fnmatch_lines(["success !"])
    assert result.ret == 0


def test_eval_comma_anycheck(testdir, extra):
    result = testdir.run("holdup", 'path://whatever123,eval://os.path.join("foo", "bar")', *extra)
    if extra:
        result.stdout.fnmatch_lines(["success !"])
    assert result.ret == 0


@pytest.mark.skipif(platform.system() != "Linux", reason="too complicated to install psycopg on non-linux")
@pytest.mark.parametrize("proto", ["postgresql", "postgres", "pg"])
def test_pg_timeout(testdir, proto):
    result = testdir.run("holdup", "-t", "0.1", proto + "://0.0.0.0/foo")
    result.stderr.fnmatch_lines(
        [
            "holdup: Failed checks: 'postgresql://0.0.0.0/foo' -> *",
        ]
    )


@pytest.mark.parametrize("proto", ["postgresql", "postgres", "pg"])
def test_pg_unavailable(testdir, proto):
    testdir.tmpdir.join("psycopg2cffi").ensure(dir=1)
    testdir.tmpdir.join("psycopg2cffi/__init__.py").write('raise ImportError("Disabled for testing")')
    testdir.tmpdir.join("psycopg2").ensure(dir=1)
    testdir.tmpdir.join("psycopg2/__init__.py").write('raise ImportError("Disabled for testing")')
    testdir.tmpdir.join("psycopg").ensure(dir=1)
    testdir.tmpdir.join("psycopg/__init__.py").write('raise ImportError("Disabled for testing")')
    result = testdir.run("holdup", "-t", "0.1", proto + ":///")
    result.stderr.fnmatch_lines(
        [
            f"holdup: error: argument service: Protocol {proto} unusable. Install holdup[[]pg[]].",
        ]
    )


@pytest.fixture
def testdir2(testdir):
    os.chdir(os.path.dirname(__file__))
    testdir.tmpdir.join("stderr").mksymlinkto(testdir.tmpdir.join("stdout"))
    return testdir


@pytest.mark.skipif("not has_docker()")
def test_func_pg(testdir2):
    result = testdir2.run("./test_pg.sh", "holdup", "pg://app:app@pg/app", "--")
    result.stdout.fnmatch_lines(["success !"])
    assert result.ret == 0


@pytest.mark.skipif("not has_docker()")
def test_func_pg_tcp_service_failure(testdir2):
    # test that the tcp check is worse than the pg check
    result = testdir2.run("./test_pg.sh", "holdup", "tcp://pg:5432", "-T", "0.001", "-i", "0", "-t", "1", "-v", "--")
    assert result.ret == 1



#!/usr/bin/python3 -u
if __name__ == "__main__":
    import sys
    import time
    import traceback

    import psycopg2

    start = time.time()
    failed = False

    while time.time() - start < 10:
        print("Connecting ...")
        try:
            conn = psycopg2.connect("postgresql://app:app@pg:5432/app")
            cursor = conn.cursor()
            cursor.execute("SELECT version()")
            print(cursor.fetchone())
        except Exception:
            traceback.print_exc()
            failed = True
        else:
            print("Connected!")
            break

    if failed:
        print("Failed!")
        sys.exit(1)
